# == Schema Information
#
# Table name: o_auths
#
#  id                   :bigint           not null, primary key
#  access_token         :string(4000)
#  account_identifier   :string(4000)
#  expires_at           :datetime
#  last_refresh_message :string(4000)
#  last_refresh_success :boolean
#  last_refresh_time    :datetime
#  raw_refresh_response :string(4000)
#  refresh_token        :string(4000)
#  refresh_url          :string(4000)
#  selected             :boolean
#  service              :integer
#  service_name         :string(4000)
#  created_at           :datetime         not null
#  updated_at           :datetime         not null
#  user_id              :bigint           not null
#
# Indexes
#
#  index_o_auths_on_user_id  (user_id)
#
# Foreign Keys
#
#  fk_rails_...  (user_id => users.id) ON DELETE => cascade
#
class OAuth < ApplicationRecord
  enum service: [:rxuniverse]
  belongs_to :user, touch: true

  compute_on_save(:selected) do |auth|
    if !auth.other_tokens_for_same_user_service.where(selected: true).exists?
      true
    else
      auth.selected || false
    end
  end

  compute_on_save(:service_name) do |auth|
    case auth.service
    when "rxuniverse"
      "Rx-Universe"
    end
  end

  compute_on_save(:refresh_url) do |auth|
    auth.refresh_url || auth.default_refresh_url
  end

  after_save do |auth|
    if auth.selected
      auth.other_tokens_for_same_user_service.update_all(selected: false)
    end
    unless auth.account_identifier.blank?
      collisions = auth.other_tokens_for_same_user_service.where(account_identifier: auth.account_identifier)
      collisions.delete_all
    end
  end

  after_destroy do |auth|
    if auth.selected
      auth.other_tokens_for_same_user_service.first&.update_columns(selected: true)
    end
  end

  scope :selected, -> do
    where(selected: true)
  end

  class << self
    def load_from_text_resp(user, service, resp_str)
      oauth = begin
        parsed = JSON.parse(resp_str)
        account_id = parsed["account_identifier"]
        OAuth.find_or_create_by(user: user, service: service, account_identifier: account_id)
      rescue
        OAuth.create(user: user, service: service)
      end
      oauth.load_from_text_resp(resp_str)
    end
  end

  def default_refresh_url
    case service
    when "rxuniverse"
      RxUniverseApiComponent.first&.oauth2_refresh_url
    end
  end

  def activate!
    update(selected: true)
  end

  def other_tokens_for_same_user_service
    OAuth.where.not(id: id).where(user: user, service: service)
  end

  def current_token!
    if expires_at && expires_at < (DateTime.now + 20.seconds)
      refresh_token!
    end
    access_token
  end

  def load_from_text_resp(resp_str)
    now = DateTime.now
    update(last_refresh_time: now, raw_refresh_response: resp_str)
    json = JSON.parse(resp_str)
    if json["access_token"].blank?
      update(
        last_refresh_message: "No access token received",
        last_refresh_success: true
      )
    else
      update(
        access_token: json["access_token"],
        refresh_token: json["refresh_token"],
        expires_at: now + json["expires_in"].to_i.seconds,
        account_identifier: json["account_identifier"],
        last_refresh_message: nil,
        last_refresh_success: true
      )
    end
  rescue => e
    update(last_refresh_message: e.to_s,
      last_refresh_success: false)
  end

  def refresh_token!
    conn = Faraday.new(refresh_url || default_refresh_url, ssl: {
      verify: !RxModule::Communicator.skip_ssl_validation
    })
    response = conn.post(nil, JSON.generate({code: refresh_token}))
    (raise response.body || "Connection Failure") unless response.success?
    load_from_text_resp(response.body)
  rescue => e
    update(last_refresh_message: e.to_s,
      last_refresh_success: false)
  end

  def to_json(*args)
    render_jbuilder_partial
  rescue
    super(*args)
  end
end
